/*
 * pp5x.c
 *
 * Programs PIC16C5X devices using hardware shown in UPP.PCX.  Based on
 * PIC16C5X programming specifications in Microchip data sheet DS30190D.
 *
 * Revision history:
 *
 * 16-Mar-1996: V-0.0; created
 * 14-Jul-1996: V-0.1; removed hardware stuff
 * 19-Nov-1996: V-0.2; first public release
 *
 * Copyright (C) 1996 David Tait.  All rights reserved.
 * Permission is granted to use, modify, or redistribute this software
 * so long as it is not sold or exploited for profit.
 *
 * THIS SOFTWARE IS PROVIDED AS IS AND WITHOUT WARRANTY OF ANY KIND,
 * EITHER EXPRESSED OR IMPLIED.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <time.h>
#include "pp5xhw.h"
#include "hex.h"

#define CP      8               /* config elements */
#define WDTE    4
#define RC      3
#define HS      2
#define XT      1
#define LP      0

struct {
    char *name;
    int device;
    int pins;
    int size;
} pic[] = {
    { "PP54", 54, 18, 512 },
    { "PP55", 55, 28, 512 },
    { "PP56", 56, 18, 1024 },
    { "PP57", 57, 28, 2048 },
    { NULL, 0, 0, 0 }
};


int valid = 0;                  /* true if lpt exists and has H/W */
int setcfg = 0;                 /* true if config set by command-line */
int dump = 0;                   /* dump PIC to a file if true */
int bcheck = 0;                 /* check PIC is blank if true */
int verify_only = 0;            /* verify PIC against a file if true */
int wait = 1;                   /* wait for PIC insertion */
int psize = 0;                  /* PIC memory size */
int device;                     /* PIC type */

char *pname = "PP5X";
char *does = "PIC16C5X Programmer";
char *version = "Version 0.2";
char *copyright = "Copyright (C) 1996 David Tait.";
char *email = "david.tait@man.ac.uk";


void quit(char *s)
{
    if ( s )
	fprintf(stderr,"%s: %s\n",pname,s);
    if ( valid )
	idle_mode();
    else if ( setup() != INVALID )
	idle_mode();
    cleanup();
    exit(0);
}


void usage(void)
{
   printf("%s  %s  %s\n\n",does,version,copyright);
   printf("Usage: %s  [ -lxhrwcdbv! ]  hexfile\n\n",pname);
   printf("       Config:   l = LP,   x = XT,    h = HS,    r = RC\n");
   printf("                 w = WDTE, c = code protect\n");
   printf("       Others:   d = dump, b = blank check, v = verify only, ");
   printf("! = no wait\n");
   printf("       Defaults: RC, /WDTE, not protected, wait \n\n");
   printf("Bug reports to %s\n",email);
   quit(NULL);
}


void get_option(char *s)
{
   char *sp;

   for ( sp=s; *sp; ++sp )
      switch ( *sp ) {
	 case 'l':
	 case 'L': config = (config&0x1C) + LP; ++setcfg; break;
	 case 'x':
	 case 'X': config = (config&0x1C) + XT; ++setcfg; break;
	 case 'h':
	 case 'H': config = (config&0x1C) + HS; ++setcfg; break;
	 case 'r':
	 case 'R': config = (config&0x1C) + RC; ++setcfg; break;
	 case 'w':
	 case 'W': config |= WDTE; ++setcfg; break;
	 case 'c':
	 case 'C': config &= ~CP; ++setcfg; break;
	 case '!': wait = 0; break;
	 case 'b':
	 case 'B': bcheck = 1; break;
	 case 'd':
	 case 'D': dump = 1; break;
	 case 'v':
	 case 'V': verify_only = 1; break;
	 case '-':
	 case '/': break;
	 default: usage();
      }
}


int read_all(void)
{
    int i;

    pmlast = -1;
    id = 0;
    cf = 0;

    prog_mode();

    if ( (config = read_cycle() & 0xF) != 0xF )
	cf = 0xFFF;

    for ( i=0; i<psize; ++i )
	if ( (progbuf[i] = read_cycle() & 0xFFF) != 0xFFF )
	    pmlast = i;

    for ( i=0; i<4; ++i )
	if ( (idbuf[i] = read_cycle() & 0xFFF) != 0xFFF )
	    id = psize;

    return !(pmlast == -1 && id == 0 && cf == 0 );
}


int blank(void)
{
    int i;

    prog_mode();

    for ( i=-1; i<psize+4; ++i )
	if ( (read_cycle() & 0xFFF) != 0xFFF )
	    return 0;
    return 1;
}


void verify(int address, U16 w, U16 r, U16 mask)
{
    if ( (w&mask) != (r&mask) ) {
	fprintf(stderr,"%s: %04X: read %03X, wanted %03X\n",
						 pname,address,r,w);
	quit("Verify failed");
    }
}


void verify_all(void)
{
    int i;

    prog_mode();

    verify(0xFFF, config, read_cycle(), 0xF);

    for ( i=0; i<psize; ++i )
	verify(i, progbuf[i], read_cycle(), 0xFFF);

    if ( id )
	for ( i=0; i<4; ++i)
	    verify(psize+i, idbuf[i], read_cycle(), 0xFFF);
}


void program(int address, U16 w)
{
    int n = 0;
    U16 r;

    if ( address == 0xFFF ) {
	if ( ((r = prog_cycle(100,w))&0xF) != (w&0xF) )
	    quit("Config failed to program");
	return;
    }

    while ( ((r = prog_cycle(1, w)) != w) && n < 25 )
	++n;

    if ( n >= 25 ) {
	fprintf(stderr,"%s: Location %04X: read %03X, wanted %03X\n"
				,pname,address,r,w);
	quit("Failed to program after 25 cycles");
    }

    prog_cycle(min(3*n,25), w);  /* over-program location */
}


void program_all(void)
{
    int i;

    prog_mode();

    read_cycle();       /* skip config word */

    for ( i=0; i<psize; ++i ) {
	program(i, progbuf[i]);
	read_cycle();
    }

    if ( id )
	for ( i=0; i<4; ++i ) {
	    program(psize+i, idbuf[i]);
	    read_cycle();
	}
}


char *conf_str(U16 cfg)
{
    static char s[4];

    s[0] = (cfg&CP)? '-': 'C';
    s[1] = (cfg&WDTE)? 'W': '-';
    s[2] = "LXHR"[cfg&3];
    s[3] = '\0';

    return s;
}

 /*
  * progname()
  *
  * Given "C:\PATH\PP54.EXE" return "PP54"
  *
  */

char *progname(char *s)
{
    static char name[5];
    char *sp = s;
    int i=0, j;

    while ( *sp++ )
      ++i;

    for ( j=0; j<4; ++j )
      name[j] = s[j+i-8];
    name[4] = '\0';

    return name;
}


void main(int argc, char *argv[])
{
   FILE *fp;
   int i, c, e, got_file = 0, t;
   char *fn;
   U16 temp;
   time_t start_t;
   int npins;

   for ( i=0; pic[i].size; ++i )
      if ( strcmp(pic[i].name, progname(argv[0])) == 0 ) {
	  pname = pic[i].name;
	  device = pic[i].device;
	  psize = pic[i].size;
	  npins = pic[i].pins;
	  break;
      }

   if ( psize == 0 )
       quit("Rename PP5X.EXE as PPXX.EXE where XX is 54, 55, 56 or 57");

   erasehex(psize, 0, 12);          /* initialise buffers */
   config = RC|CP;             	    /* default config */

   for ( i=1; i<argc; ++i ) {
      if ( (c = *argv[i]) == '-' || c == '/' )
	 get_option(++argv[i]);
      else {
	 if ( got_file )
	     usage();
	 fn = argv[i];
	 got_file = 1;
      }
   }

   switch ( setup() ) {
       case INVALID:
	   quit("Illegal LPT port (check PPLPT setting)");
       case NOLOOP:
	   quit("Hardware not found");
       case BADREGS:
	   quit("Hardware fault");
   }

   valid = 1;

   if ( one_socket ) {
       if ( has18pins ) {
	   if ( npins == 28 )
	       quit("Hardware configured for 18 pin PICs");
       } else if ( npins == 18 )
	    quit("Hardware configured for 28 pin PICs");
   }

   if ( !bcheck && !got_file )             /* no file needed for bcheck */
       usage();

   if ( dump && !got_file )                /* need file for dump */
       usage();

   if ( verify_only && !got_file )         /* need file for verify */
       usage();

   printf("%s  %s  %s\n\n",does,version,copyright);

   if ( wait ) {
      printf("Insert PIC and press a key to start (^C to abort) ...\n\n");
      if ( getch() == 3 )
	 quit("Aborted");
   }

    if ( dump ) {
       if ( read_all() ) {
	   if ( (fp = fopen(fn,"w")) == NULL )
	       quit("Can't create hexfile");
	   format = INHX8M;
	   dumphex(fp, pmlast+1, 0, psize, 0xFFF, 0);
	   fclose(fp);
	   quit(NULL);
       }
       quit("PIC is blank - nothing to dump");
    }

    if ( got_file ) {
	if ( (fp = fopen(fn,"r")) == NULL )
	    quit("Can't open hexfile");
	temp = config;
	if ( (e = loadhex(fp, psize, 0, psize, 0xFFF, 0)) < 0 )
	    quit(errhex(e));
	if ( setcfg )           /* use command-line config if set */
	    config = temp;
	else if ( cf > 0 )
	    ++setcfg;
    }

    start_t = time(NULL);
    if ( verify_only ) {
	printf("Verifying ...\n");
	verify_all();
    } else {
       if ( bcheck ) {
	   printf("Checking ...\n");
	   if ( !blank() )
	      quit("PIC is not blank");
	   if ( !got_file )
	       quit("PIC is blank\n");
       }
       printf("Programming ...\n");
       program_all();
       printf("Programming config to %s ... %s\n",
	       conf_str(config), setcfg? "": "(Warning: using default!)");
       prog_mode();
       program(0xFFF, config);
   }
   t = (int) (time(NULL)-start_t);
   printf("Finished in %d sec%c\n\n",t,(t!=1)? 's': ' ');
   quit(NULL);
}
