/*
 * pp5xhw.c
 *
 * Hardware level routines used by the PP5X programmer.  Information
 * about the parallel port was taken from Zhahai Stewart's PC printer
 * port FAQ.
 *
 * Revision history:
 *
 * 16-Mar-1996: V-0.0; started as part of pp5x.c
 * 14-Jul-1996: V-0.1; created separate file
 * 19-Nov-1996: V-0.2; first public release
 *
 * Copyright (C) 1996 David Tait.  All rights reserved.
 * Permission is granted to use, modify, or redistribute this software
 * so long as it is not sold or exploited for profit.
 *
 * THIS SOFTWARE IS PROVIDED AS IS AND WITHOUT WARRANTY OF ANY KIND,
 * EITHER EXPRESSED OR IMPLIED.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "pp5xhw.h"

static char *version = "pp5xhw.c V-0.2  Copyright (C) 1996 David Tait";

int d_reg;             /* port address of LPT data register */
int s_reg;             /* LPT status register */
int c_reg;             /* LPT control register */

int d_bits;            /* data port mirror */
int c_bits;            /* control port mirror */

int tclk = TCLK;       /* default register clock pulse width in ticks */

int one_socket = 1;    /* true if one socket used for 18 and 28 pin PICs */
int has18pins;         /* non-zero for 18 pin setup */


  /*
   * read_reg(), fill_reg()
   *
   * Routines to read/write 12-bit words via two shift registers.  This
   * is how data is transferred to/from 16C5X chips.  If the hardware is
   * designed so that both 18 and 28 pin PICs are programmed using a
   * common socket then the data bits change significance depending on
   * the current type - the routines must know the hardware setup and
   * scramble the data accordingly.
   *
   */

static U16 read_reg(void)
{
    int bit, s, wh, wl, w;

    ENABLE();           /* can only shift while outputs are enabled */

    wh = 0;
    wl = 0;
    for ( bit=0; bit<8; ++bit) {
	s = INS();
	wh = (wh<<1) + ((s&S_RHO)? 1: 0);
	wl = (wl<<1) + ((s&S_RLO)? 1: 0);
	us_pulse(RCL, tclk);
    }
    w = (wh<<4) + (wl&0xF);


    if ( one_socket && has18pins )    /* unscramble data */
	w = ((w&3)<<2) + ((w&0xC)>>2) + ((w&0xFE0)>>1) + ((w&0x10)<<7);

    DISABLE();
    return w;
}


static void fill_reg(U16 w)
{
   int bit, wh, wl;

   ENABLE();            /* can only shift while outputs are enabled */

   if ( one_socket && has18pins )    /* scramble data */
	w = ((w&3)<<2) + ((w&0xC)>>2) + ((w&0x7F0)<<1) + ((w&0x800)>>7);

   wl = w&0xF;
   wh = (w&0xFF0)>>4;

   for ( bit=0; bit<8; ++bit) {
	(wh&BIT(7))? RHI_HI(): RHI_LO();
	(wl&BIT(7))? RLI_HI(): RLI_LO();
	us_pulse(RCL, tclk);
	wh <<= 1;
	wl <<= 1;
   }

   DISABLE();
}


void idle_mode(void)
{
    IDLE();     /* all LPT data pins high */
    RESET();    /* /MCLR low */
    ms_delay(TVDD);
}


void prog_mode(void)
{
    idle_mode();
    SDA_HI();           /* /PROG/VER = 1 */
    SCL_LO();           /* INCPC = 0 */
    OUTPUT();
    ON();
    ms_delay(TVDD);
    PROGRAM();
    us_delay(TSET);
}


  /*
   * prog_cycle(n, w)
   *
   * Programs a location n times with word w reading PIC contents
   * on the last cycle.
   *
   */

U16 prog_cycle(int n, U16 w)
{
    int i;

    fill_reg(w);
    for (i = 0; i < n; ++i ) {
	ENABLE();
	us_delay(TSET);
	us_pulse(SDA, TPW);
	us_delay(THLD);
	DISABLE();
	us_delay(TREC);
	if ( i == n-1 ) {
	    SDA_LO();
	    us_pulse(RCL, tclk); /* read PIC on last cycle */
	    SDA_HI();
	} else
	    us_pulse(SDA, tclk);
	us_delay(THLD);
    }
    return read_reg();
}


  /*
   * read_cycle()
   *
   * Increment PC and read PIC contents.
   *
   */

U16 read_cycle(void)
{
    SCL_HI();
    us_pulse(RCL, tclk);
    SCL_LO();
    us_delay(THLD);
    return read_reg();
}


void cleanup(void)
{
   reset_timer();
}


  /*
   * setup()
   *
   * Does all the housekeeping required including: finding the parallel
   * port address; setting up the CTC; performing sanity checks on the
   * hardware and reading the 18/28 pin flag.
   *
   */

int setup(void)
{
    int w, lpt = 1;
    char *s;

    if ( (s = getenv("PPLPT")) != NULL )
	if ( (lpt = atoi(s)) < 1 || lpt > 3 ) 
	    lpt = 1;

    if ( (s = getenv("PPCLK")) != NULL )
	if ( (tclk = atoi(s)) < 1 || tclk > 127 )
	    tclk = TCLK;

    if ( (s = getenv("PPSOCK")) != NULL )
	if ( atoi(s) != 1 )
	    one_socket = 0;

    d_reg = *(((int far *) 0x408) + lpt-1);  /* base address of LPT port */
    s_reg = d_reg+1;
    c_reg = s_reg+1;

    switch ( d_reg ) {                   /* check port address is valid */
	case 0x3BC:
	case 0x378:
	case 0x278: break;
	   default: return INVALID;
    }

    idle_mode();
    ON();       /* power on */
    OUTPUT();   /* SDA is an output */

    for ( w=0; w<4; ++w) {              /* loopback test */
	(w&2)? SDA_HI(): SDA_LO();
	(w&1)? SCL_HI(): SCL_LO();
	if ( w != (INS()>>6) )
	    return NOLOOP;
    }

    fill_reg(0xABC);                    /* register test */
    if ( read_reg() != 0xABC )
	return BADREGS;

    has18pins = (INS()&S_28P)? 0: 1;

    idle_mode();
    return OK;
}

